--[[
	LevelLogger

	This addon records a log of how long it took to reach each level.
	Use /lvlog to print the current character's log in the chat frame.
	With no paramters the log report window will be shown instead.

	The following parameters are optional:
		/lvlog <number> <level> <channel> <who>
		number = reports the last <number> levels up until <level>, default = 5
			use zero to report all levels logged
		level = reports the log up until this level, default = current level
		channel = indicates a chat channel to report in, default = console output
			s = /say
			y = /yell
			p = /party
			r = /raid
			g = /guild
			w = /whisper
		who = if using channel w, this specifies the whisper target, default = current target
]]--

local L = LibStub("AceLocale-3.0"):GetLocale("LevelLogger");
local AceGUI = LibStub("AceGUI-3.0");
local gTimePlayedPending = false;

LevelLogger = LibStub("AceAddon-3.0"):NewAddon( "LevelLogger",
	"AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0"
);

-- saved variable containing the level log for each character
svLevelLog = {};

-- reporting options, saved globally
svLevelLogOptions = {
	showCurrDate = false,
	showCurrTime = false,
	showLevelTime = true,
	showPlayedTime = true,
	reportToParty = true,
	reportToGuild = true,
};

-- similar to "Fill" layout, except there is a header and footer child frame
-- and the middle child "body" frame fills up the remainder of the content frame
-- if the "body" frame is a ScrollFrame, the header will auto-adjust width for the scrollbar
AceGUI:RegisterLayout("Table",
	function(content, children)
		if ( #children < 2 ) then return; end

		-- show the frames
		local header = children[1];
		local body = children[2];
		local footer = children[3];
		header.frame:Show();
		body.frame:Show();
		if (footer) then footer.frame:Show(); end

		-- position the frames
		local width = content.width or content:GetWidth() or 0;
		local height = content:GetHeight() or 0;
		header.frame:SetPoint("TOPLEFT", content);
		body.frame:SetPoint("TOPLEFT", header.frame, "BOTTOMLEFT");
		if (footer) then
			footer.frame:SetPoint("BOTTOMLEFT", content);
			footer.frame:SetPoint("BOTTOMRIGHT", content);
			body.frame:SetPoint("BOTTOMLEFT", footer.frame, "TOPLEFT");
		else
			body.frame:SetPoint("BOTTOMLEFT", content);
		end
		body:SetWidth( width );
		header:SetWidth( width - (body.scrollBarShown and 20 or 0) );

		-- finish the layout
		if content.obj.LayoutFinished then
			content.obj.LayoutFinished( content.obj, nil, height);
		end

		--print ( "DoLayout: width=" .. width .. "; height=" .. height .. "; scroll=" .. (body.scrollBarShown and "yes" or "no") );
		--print( "DoLayout: s_height=" .. body.scrollframe:GetHeight() .. "; s_viewheight=" .. body.content:GetHeight() );
	end)

function LevelLogger:OnInitialize()
	-- init workspace vars
	self.newLevel = false;		-- gets set on ding, then used when /played is processed
	self.totalTime = 0;			-- gets set on ding, then used when /played is processed
	self.levelTime = 0;			-- gets set when /played is processed
	self.showCurrent = false;	-- used by "current" command to show time played to current level
	self.startup = true;		-- true between addon loaded and the first /played
	self.playerName = UnitName("player");

	-- init stuff
	self:SetupOptions();
	self:RegisterChatCommand( "lvlog", "OnChatCommand" );

	-- register events
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("PLAYER_LEVEL_UP");
	self:RegisterEvent("TIME_PLAYED_MSG");

	-- hook the /played event to prevent it from being called multiple times by other addons
	self:RawHook( "RequestTimePlayed", true );

	--print ( "LevelLogger init complete" );
end

function LevelLogger:SetupOptions()
	LibStub("AceConfig-3.0"):RegisterOptionsTable( "LevelLogger", {
		type = "group",
		get = function(info) return svLevelLogOptions[info[#info]]; end,
		set = function(info,val) svLevelLogOptions[info[#info]] = val; end,
		args = {
			showCurrDate = {
				name = L["Report Level-Up Date"],
				desc = L["Report the date when each level is logged"],
				type = "toggle",
				width = "full",
			},
			showCurrTime = {
				name = L["Report Level-Up Time"],
				desc = L["Report the time of day when each level is logged"],
				type = "toggle",
				width = "full",
			},
			showLevelTime = {
				name = L["Report Time Played to Level-Up"],
				desc = L["Report the time /played to reach each level"],
				type = "toggle",
				width = "full",
			},
			showPlayedTime = {
				name = L["Report Total Time Played"],
				desc = L["Report the total time /played when each level is logged"],
				type = "toggle",
				width = "full",
			},
			reportToParty = {
				name = L["Send Level-Up Report to Party"],
				desc = L["Send level up messages to your party's chat channel"],
				type = "toggle",
				width = "full",
			},
			reportToGuild = {
				name = L["Send Level-Up Report to Guild"],
				desc = L["Send level up messages to your guild's chat channel"],
				type = "toggle",
				width = "full",
			},
		}
	} );
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions( "LevelLogger" );
	self.optionsFrame.default = function() self:SetDefaultOptions(); end;
	self.optionsReg = LibStub("AceConfigRegistry-3.0");
end

function LevelLogger:SetDefaultOptions()
	svLevelLogOptions.showCurrDate = false;
	svLevelLogOptions.showCurrTime = false;
	svLevelLogOptions.showLevelTime = true;
	svLevelLogOptions.showPlayedTime = true;
	svLevelLogOptions.reportToParty = true;
	svLevelLogOptions.reportToGuild = true;

	--force config dialog to show new state of options
	self.optionsReg:NotifyChange("LevelLogger");
end

-- returns logData record formatted into text string
function LevelLogger:LogFormat( logData )
	local strLog = format( L["Dinged level %d"], logData.level );
	if (logData.ts ~= nil) then
		if (svLevelLogOptions.showCurrDate) then
			if (svLevelLogOptions.showCurrTime) then
				-- date and time
				strLog = strLog..date( L[" on %d-%b-%Y %H:%M:%S"], logData.ts );
			else
				-- date only
				strLog = strLog..date( L[" on %d-%b-%Y"], logData.ts );
			end
		elseif (svLevelLogOptions.showCurrTime) then
			-- time only
			strLog = strLog..date( L[" at %H:%M:%S"], logData.ts );
		end
	end

	if (logData.levelTime > 0 and svLevelLogOptions.showLevelTime and svLevelLogOptions.showPlayedTime) then
		local d1, h1, m1, s1 = ChatFrame_TimeBreakDown(logData.levelTime);
		local d2, h2, m2, s2 = ChatFrame_TimeBreakDown(logData.totalTime);
		strLog = strLog..format( L[" in %d days %02d:%02d:%02d (of %d days %02d:%02d:%02d total played)"],
			d1, h1, m1, s1, d2, h2, m2, s2);
	elseif (logData.levelTime > 0 and svLevelLogOptions.showLevelTime) then		-- include time to level with ding
		local d, h, m, s = ChatFrame_TimeBreakDown(logData.levelTime);
		strLog = strLog..format( L[" in %d days %02d:%02d:%02d"], d, h, m, s);
	elseif (svLevelLogOptions.showPlayedTime) then		-- include /played time with ding
		local d, h, m, s = ChatFrame_TimeBreakDown(logData.totalTime);
		strLog = strLog..format( L[" at %d days %02d:%02d:%02d total played"], d, h, m, s);
	end

	return strLog;
end

-- appends logData fields into a table of report columns
function LevelLogger:LogEntry( logData, out )
	if ( out.level == nil ) then
		-- init the table columns
		out.level = {};
		out.date = {};
		out.levelTime = {};
		out.totalTime = {};
	end

	out.level[ #out.level + 1 ] = logData.level;
	if (logData.ts ~= nil) then
		out.date[ #out.date + 1 ] = date( L["%d-%b-%Y %H:%M:%S"], logData.ts );
	else
		out.date[ #out.date + 1 ] = "";
	end
	if (logData.levelTime > 0) then
		local d, h, m, s = ChatFrame_TimeBreakDown(logData.levelTime);
		out.levelTime[ #out.levelTime + 1 ] = format( L["%d days %02d:%02d:%02d"], d, h, m, s);
	else
		out.levelTime[ #out.levelTime + 1 ] = "";
	end
	if (svLevelLogOptions.showPlayedTime) then
		local d, h, m, s = ChatFrame_TimeBreakDown(logData.totalTime);
		out.totalTime[ #out.totalTime + 1 ] = format( L["%d days %02d:%02d:%02d"], d, h, m, s);
	else
		out.totalTime[ #out.totalTime + 1 ] = "";
	end
end

function LevelLogger:ShowReport()
	-- build the log report table
	local report = {};
	for i,logData in pairs(svLevelLog) do
		self:LogEntry( logData, report );
	end

	-- create the report window
	local g, h, t, s;
	g = AceGUI:Create("Window");
	g:SetTitle( format( L["LevelLogger report for %s"], self.playerName ) );
	g:SetCallback("OnClose", function(w) AceGUI:Release(w); end);
	g:SetLayout("Table");
	g:SetWidth(500);
	g:SetHeight(350);
	g.frame:SetMinResize(420, 240);
	g.frame:SetBackdrop( {
		bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
		edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
		tile = true, tileSize = 32, edgeSize = 32,
		insets = { left = 8, right = 8, top = 8, bottom = 8 }
	} );
	g.frame:SetBackdropColor( 0,0,0,1 );

	-- create the header
	h = AceGUI:Create("SimpleGroup");
	h:SetLayout("Flow");
	g:AddChild(h);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.102); t:SetText( L["Level"] );
	h:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.372); t:SetText( L["Date"] );
	h:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.262); t:SetText( L["Level Time"] );
	h:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.262); t:SetText( L["Total Time"] );
	h:AddChild(t);

	-- create the scroll frame
	s = AceGUI:Create("ScrollFrame");
	s:SetLayout("Flow");
	g:AddChild(s);

	-- populate the column text
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.102); t:SetText( table.concat( report.level, "\n" ) );
	s:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.372); t:SetText( table.concat( report.date, "\n" ) );
	s:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.262); t:SetText( table.concat( report.levelTime, "\n" ) );
	s:AddChild(t);
	t = AceGUI:Create("Label");
	t:SetRelativeWidth(0.262); t:SetText( table.concat( report.totalTime, "\n" ) );
	s:AddChild(t);

	-- display the report
	s:FixScroll(); s:FixScroll(); -- need to call twice to know if scroll bar is initially visible
	g:DoLayout();
	--self:Print( "ShowReport: s_height=" .. s.scrollframe:GetHeight() .. "; s_viewheight=" .. s.content:GetHeight() .. "; scroll=" .. (s.scrollBarShown and "yes" or "no") );

	if ( s.scrollBarShown ) then s.scrollbar:SetValue(1000); end
end

function LevelLogger:OnChatCommand(msg)
	if (msg == "") then
		-- show the report window if no arguments
		self:ShowReport();
		return;
	end

	local args = { self:GetArgs( msg, 4 ) };
	if (args[1] == "help" or args[1] == "?" or args[1] == "h") then
		-- display the help text
		for i,txt in pairs( L["helpText"] ) do
			print( txt );
		end
	elseif (args[1] == "current") then
		-- calculate and log time to current level (in case it was missed)
		self.showCurrent = true;
		RequestTimePlayed();
--	elseif (args[1] == "testding") then
--		self:TestDing();
	elseif (#svLevelLog == 0) then
		self:Print( format( L["No LevelLogger data found for %s"], self.playerName ) );
	else
		local numLevels = tonumber( args[1] );
		local toLevel = tonumber( args[2] );
		local channel = args[3];
		if (numLevels == nil) then
			numLevels = 5; -- show last 5 levels by default
		elseif (numLevels < 1) then
			numLevels = 999; -- show all levels
		end
		local myLevel = UnitLevel("player");
		if (toLevel == nil or toLevel < 1 or toLevel > myLevel) then
			toLevel = myLevel; -- show most recent levels
		end
		local logFunc;
		if (channel ~= nil) then
			local ch = channel:sub(1,1):lower();
			local map = { s = "SAY", y = "YELL", p = "PARTY", r = "RAID", g = "GUILD", w = "WHISPER" };
			if ( map[ch] ~= nil ) then
				if ( ch ~= "w" ) then
					local okay = true;
					if ( ch == "p" ) then
						okay = GetNumSubgroupMembers() > 0;
					elseif ( ch == "r" ) then
						okay = GetNumGroupMembers() > 0;
					elseif ( ch == "g" ) then
						okay = IsInGuild();
					end
					if ( okay ) then
						logFunc = function( txt ) SendChatMessage( txt, map[ch] ); end
					end
				else -- get the whisper target
					local w = args[4];
					if ( w == nil or w == "%t" ) then
						w = UnitName("target");
					elseif ( not UnitIsPlayer( w ) ) then
						w = nil;
					end
					if ( w ~= nil ) then
						logFunc = function( txt ) SendChatMessage( txt, "WHISPER", nil, w ); end
					end
				end
			end
		end
		if ( logFunc == nil ) then
			logFunc = function( txt ) self:Print( txt ); end
		end
		-- print out the level log
		logFunc( format( L["LevelLogger report for %s"]..":", self.playerName ) );
		for i,logData in pairs(svLevelLog) do
			local lvl = tonumber(logData.level);
			if (lvl > toLevel - numLevels and lvl <= toLevel) then
				logFunc( self:LogFormat( logData ) );
			end
		end
	end
end

-- combine multiple pending /played messages into just one
function LevelLogger:RequestTimePlayed()
	if not gTimePlayedPending then
		gTimePlayedPending = true;
		self.hooks["RequestTimePlayed"]();
	end
end

function LevelLogger:TestDing()
	self.newLevel = 99;
	RequestTimePlayed();
end

function LevelLogger:Ding(lvl,totalTime,levelTime)
	if ( lvl < 2 ) then return; end

	local logData = {
		level = tonumber(lvl);
		levelTime = levelTime;
		totalTime = totalTime;
		ts = self.newLevel and time();
	};

	local lastLog, lastLevel;
	if ( self.showCurrent ) then
		-- check if this level is already logged
		if ( #svLevelLog > 0 ) then
			lastLog = svLevelLog[ #svLevelLog ];
			lastLevel = lastLog.level;
			if ( lastLog.level == lvl and #svLevelLog > 1 ) then
				-- get record for previous level
				lastLog = svLevelLog[ #svLevelLog - 1 ];
			end
			if ( lastLog.level == lvl - 1 ) then
				-- calculate levelTime based on log from previous level
				logData.levelTime = logData.totalTime - lastLog.totalTime;
			end
		end
		if ( lvl == 2 and logData.levelTime == 0 ) then
			-- specially handle level 2 since there are no prior levels
			logData.levelTime = logData.totalTime;
		end
	end
	if ( not lastLevel or lastLevel < lvl ) then
		-- append a new level to the log
		table.insert( svLevelLog, logData );
	end

	local strLog = self:LogFormat( logData );
	local sentmsg = false;
	if ( self.newLevel ) then
		if ( svLevelLogOptions.reportToParty and GetNumSubgroupMembers() > 0 ) then
			SendChatMessage( strLog, "PARTY" );
			sentmsg = true;
		end
		if ( svLevelLogOptions.reportToGuild and IsInGuild() ) then
			SendChatMessage( strLog, "GUILD" );
			sentmsg = true;
		end
	end
	if ( not sentmsg ) then
		-- print out the text if not already displayed
		self:Print(strLog);
	end
end

function LevelLogger:PLAYER_ENTERING_WORLD()
	--self:Print( "[PLAYER_ENTERING_WORLD] totalTime=" .. (self.totalTime or "nil") .. "; levelTime=" .. (self.levelTime or "nil") );
	-- always update the time played when zoning
	RequestTimePlayed();
end

function LevelLogger:PLAYER_LEVEL_UP(eventname, theLevel)
	self.newLevel = theLevel;
	RequestTimePlayed();
end

function LevelLogger:TIME_PLAYED_MSG(eventname, timePlayedTotal, timePlayedLevel)
	gTimePlayedPending = false;
	if (self.newLevel) then
		-- timePlayedLevel should be zero, so we need to calculate the level up time
		-- use last known times and add the difference from current total played time
		local levelTime = self.levelTime + ( timePlayedTotal - self.totalTime );
		self:Ding(self.newLevel,timePlayedTotal,levelTime);
		self.newLevel = false;
	elseif (self.showCurrent) then
		self:Ding(UnitLevel("player"), timePlayedTotal - timePlayedLevel, 0);
		self.showCurrent = false;
	end
	self.totalTime = timePlayedTotal;
	self.levelTime = timePlayedLevel;
	--self:Print( "[TIME_PLAYED_MSG] totalTime=" .. timePlayedTotal .. "; levelTime=" .. timePlayedLevel );
end
